
      { cloak.pas, 90/10/26/tfr (from 10/23, 9/3, 8/31-30,9,4-1,
                                      7/31-28,3-2, 6/30-29,16-15) }
      { Double Dynamic Substitution cipher }
      { Turbo Pascal 5.5 }

      {    Terry Ritter        }
      {    Blue Jean Software  }
      {    2609 Choctaw Trail  }
      {    Austin, TX  78745   }
      {    (512) 892-0494      }

      { (c) Copyright 1990, T. F. Ritter; All Rights Reserved }
      { Patent Pending }

      { This source file and the resulting .EXE program }
      {    may be freely copied, but only if unmodified. }
      { Modified versions may not be publicly distributed. }


      { project 211 }



      { This program is a DOS "filter": it ciphers StdIn to StdOut. }
      {    This allows the smallest practical program source. }

      { The usual command line is:  CLOAK <fromfile >tofile }

      { The enciphering/deciphering decision is normally automatic }
      {    based on a value at the start of enciphered files. }
      {    Use option "/e" to re-encipher enciphered files. }
      { Force Enciphering:  CLOAK /E <fromfile >tofile }

      { Initialization delay is partly related to cipher strength, and }
      {    can be reduced with option "/l" (Low level = deg 1279) }
      {    or increased with option "/h" (High level = deg 11213). }
      {    The normal cipher level is deg 4253. }
      { Faster:    CLOAK /L <fromfile >tofile }
      { Stronger:  CLOAK /H <fromfile >tofile }




{ $ DEFINE Debug }  { enable slow stack/range checks }
{$DEFINE InlRNG }  { enable fast inline RNG }
{$DEFINE InlExg }  { enable fast inline Exchange Char }



      PROGRAM cloak;


      {$A+}   { Word Alignment ON }
      {$B-}   { Full Boolean Evaluation OFF }
      {$D+}   { Debug Information ON }
      {$F-}   { Far Calls OFF }
      {$I-}   { I/O-Checking OFF }
      {$L+}   { Local Symbols ON }
      {$N-}   { Numeric Co-Processor OFF }
      {$O-}   { Overlay Code OFF }
      {$V-}   { VAR-String Checking OFF }

{$IFDEF Debug}
      {$R+}   { Range-Checking ON }
      {$S+}   { Stack-Checking ON }
{$ELSE}
      {$R-}   { Range-Checking OFF }
      {$S-}   { Stack-Checking OFF }
{$ENDIF}




      USES
         Crt;  { KeyPressed, Readkey, Output }



      PROCEDURE Ownership;
         BEGIN
         WRITELN( Output,
^M^J^J'  CLOAK  File Ciphering Filter ͻ ',
  ^M^J'  Syntax:  cloak[opt] <fromfile >tofile   ',
  ^M^J'  Example:  cloak/o -- show options       ',
  ^M^J' Ķ ',
  ^M^J'  CLOAK is a FILTER!  Use command-line    ',
  ^M^J'   redirection, or abort with Ctrl-Z.     ',
  ^M^J'  Copyright 1990, T.F. Ritter ͼ ' );
         END; {ownership}



      PROCEDURE ShowOptions;
         BEGIN
         WRITELN( Output,
  ^M^J'  CLOAK Options ͻ ',
  ^M^J'  RNG Level  (norm = deg 4253) Ķ ',
  ^M^J'     /l -- Low = deg 1279                 ',
  ^M^J'     /h -- High = deg 11213               ',
  ^M^J'  Mode  (normally automatic) Ķ ',
  ^M^J'     /e -- force Encipher mode            ',
  ^M^J'       (allows repeated enciphering)      ',
  ^M^J' ͼ ' );
         END; {showoptions}



      PROCEDURE Advert;
         BEGIN
         WRITELN( Output,
  ^M^J'  Thank You for Using CLOAK! ͻ ',
  ^M^J'  $35 registration required if used for   ',
  ^M^J'  more than one conversation or purpose.  ',
  ^M^J' ͹ ',
  ^M^J'  Blue Jean Software      (512) 892-0494  ',
  ^M^J'  2609 Choctaw Trail    Austin, TX 78745  ',
  ^M^J'  Patent Pending ͹ ',
  ^M^J'  CLOAK may be freely copied but only if  ',
  ^M^J'  unmodified.    Ver 1.1, 90-10-26/tfr.   ',
  ^M^J' ͼ ' );
         END; {advert}



      { start user key input }


      FUNCTION GetHiddenLine( VAR s: STRING ): BOOLEAN;
         { collect chars from kbd until CR w/o screen display }
         VAR
            ch: CHAR;
            lens: BYTE ABSOLUTE s;
         BEGIN
         GetHiddenLine := FALSE;
         lens := 0;  { length }
            REPEAT
            REPEAT UNTIL KeyPressed;
            ch := ReadKey;
               CASE ch OF
         {CR}  ^M: BEGIN  GetHiddenLine := (lens > 0);  Exit;  END;  {success}
         {LF}  ^J: ;  {skip}
         {BS}  ^H: IF (lens <> 0) THEN  Dec(lens);  {backspace}
   {ESC/QUIT}  ^[: IF (lens = 0) THEN
                      Exit   {Esc at start of line Quits }
                   ELSE
                      lens := 0;  {start line over}
               ELSE {case}
                  INC(lens);  s[lens] := ch;  {save char}
               END; {case}
            UNTIL (lens = PRED(SizeOf(s)));  {full}
         GetHiddenLine := TRUE;
         END; {GetHiddenLine}



      FUNCTION GetUserKey( VAR s: STRING ): BOOLEAN;
         { get user key until same string is input twice, or quits }
         VAR
            s2: STRING;
            ok: BOOLEAN;
         BEGIN
         GetUserKey := FALSE;
            REPEAT
            WRITE( Output, ^M^J' Please enter user key, or Esc to abort: ' );
            IF NOT GetHiddenLine( s ) THEN  Exit;  { quit exit }
            WRITE( Output, ^M^J' Please enter user key again: ' );
            ok := GetHiddenLine( s2 ) AND (s = s2);
            IF ok THEN
               WRITE ( Output, ' ok. ' )
            ELSE
               WRITE( Output, ' MISMATCH! ' );
            UNTIL (ok);
         FillChar( s2, SizeOf(s2), 0 );  { hide check key }
         GetUserKey := TRUE;
         END; {GetUserKey}



      { start user key processing }


      TYPE
         Lwd = LONGINT;


      FUNCTION crc31xbyf( crc, poly: Lwd; d: BYTE ): Lwd;
         { whole byte CRC, alternate method, general deg 31 poly }
         VAR
            i: BYTE;
            crcba: ARRAY[ 0..3 ] of BYTE ABSOLUTE crc;
         BEGIN
         crc := crc Shl 1;
         crcba[3] := crcba[3] XOR d;
         IF (LONGINT(crc) < 0) THEN
            crc := crc XOR poly;
         FOR i := 6 DOWNTO 0 DO
            BEGIN
            crc := crc Shl 1;
            IF (LONGINT(crc) < 0) THEN
               crc := crc XOR poly;
            END;
         crc31xbyf := crc;
         END; {crc31xbyf}



      TYPE
         ByteArray = ARRAY[ 0..65520 ] of BYTE;


      FUNCTION ovcrc31f( crc, poly: Lwd; VAR buf; byct: WORD ): Lwd;
         { memory-scan CRC, general deg 31 poly }
         VAR
            databyar: ByteArray ABSOLUTE buf;
            i: WORD;
         BEGIN
         FOR i := 0 TO PRED(byct) DO
            crc := crc31xbyf( crc, poly, databyar[i] );
         ovcrc31f := crc;
         END; {ovcrc31f}



      TYPE
         Lwd32AR = ARRAY[ 0..31 ] of Lwd;
         LwdAR = ARRAY[ 0..16382 ] of Lwd;


      CONST  { deg 31 PRIMITIVE mod 2 polynomials for user key expansion }
         { these particular polynomials (c) Copyright 1990, T. F. Ritter }
         Poly31AR: Lwd32AR = (
            $a021affb, $ac054145, $942a011d, $a3c0420d,
            $b6b1d435, $cb866af7, $b740ef79, $ca8440a5,
            $d5134ff5, $92480167, $97254891, $8b53351f,
            $810070f5, $b27aa9af, $8465d359, $a8353c47,
            $8c12305f, $8978dc97, $86a1f4eb, $91d081b3,
            $9a2a23bf, $9bdc3e35, $d7840839, $c2116213,
            $d0c0803b, $9d0094df, $a00a64dd, $80219add,
            $84841d7f, $88376ff5, $805810db, $8248def7 );



      PROCEDURE UserKeyCRCinit( VAR init; VAR polyAR: Lwd32AR;
                                VAR buf; byct: WORD );
         { run multiple deg 31 CRC's on user key; results to init }
         VAR
            initAR: LwdAR ABSOLUTE init;
            i: BYTE;
         BEGIN

         FOR i := 0 TO 31 DO
            initAR[i] := ovcrc31f( $ffffffff, polyAR[i], buf, byct );

         END; {UserKeyCRCinit}



      PROCEDURE Pack31bitLwd32AR( VAR init );
         { pack 32 31-bit values (b30-b0) into 31 32-bit values }
         VAR
            initAR: Lwd32AR ABSOLUTE init;
            i, j: BYTE;
         BEGIN
         FOR i := 31 DOWNTO 0 DO
            { create tightly packed array element by element }
            { packed area starts at [31] and eventually ends at [0] }
            BEGIN
            initAR[i] := initAR[i] Shl 1;  { lose msb }
               { [i] will head the new tightly packed area }
            FOR j := SUCC(i) TO 31 DO  { shift w/o loss }
               { shift previous tightly packed area left one bit }
               BEGIN
               IF (LONGINT(initAR[j]) < 0) THEN  Inc( initAR[PRED(j)] );
               initAR[j] := initAR[j] Shl 1;
               END;
            END;
         END; {Pack31bitLwd32AR}


      { end user key input and processing }



      { start really-random generator }
      { based on human keystroke-timing delays }


      CONST
         TimerCtrlPort = $43;
         Timer0DataPort = $40;
         Timer0Mode2 = $34;
         Timer0Mode3 = $36;


      PROCEDURE TimerInit;
         { prepare for timing by setting mode 2 }
         BEGIN
         Port[ TimerCtrlPort] := Timer0Mode2;
         Port[ Timer0DataPort] := 0;
         Port[ Timer0DataPort] := 0;
         END; {TimerInit}


      PROCEDURE TimerRecover;
         { restore mode 3 }
         BEGIN
         Port[ TimerCtrlPort] := Timer0Mode3;
         Port[ Timer0DataPort] := 0;
         Port[ Timer0DataPort] := 0;
         END; {TimerRecover}


      PROCEDURE IoFlush; Inline( $eb/ $00 );


      FUNCTION GetCount: WORD;
         { return 16-bit timer0 count value (useful only in mode 2) }
         { this is the HIGH SPEED counter = 1.19 MHz count rate }
         { the count value rolls over 18.2 times /sec }
         VAR
           x, y: BYTE;
         BEGIN
         Port[ TimerCtrlPort ] := 0;  { LATCH THE COUNT! }
         IoFlush;
         x := Port[ Timer0DataPort ];  { ls }
         y := Port[ Timer0DataPort ];  { ms }
         GetCount := WORD( -(Swap(y) + x) );
         END; {GetCount}



      CONST  { set size of character history array }
         HistPwr = 3;   { 4 => HistMax = 15; 3 => HistMax = 7 }
         HistMax = PRED( 1 Shl HistPwr );

      VAR  { static memory allocation: character history storage }
         HistAr: ARRAY[ 0..HistMax ] of CHAR;
         HistCur: BYTE;



      FUNCTION NewChar( ch: CHAR ): BOOLEAN;
         { ignore recently-used keys; accept only NEW keys }
         {    thus avoid auto-repeat and fixed finger placement }
         VAR
            i: BYTE;
         BEGIN
         NewChar := FALSE;
         IF (ch = ^@) THEN  Exit;

         { linear search for current char }
         FOR i := 0 TO HistMax DO
            IF (ch = HistAr[i]) {found} THEN  Exit;

         { not found, so save as history }
         HistAr[ HistCur AND HistMax ] := ch;
         Inc( HistCur );

         NewChar := TRUE;
         END; {NewChar}



      CONST
         polycrc32a = $04c11db7;



      FUNCTION crc32abyfb( crc: Lwd; d: BYTE ): Lwd;
         { whole byte CRC, alternate method }
         VAR
            i: BYTE;
            crcba: ARRAY[ 0..3 ] of BYTE ABSOLUTE crc;
         BEGIN
         crcba[3] := crcba[3] XOR d;
         FOR i := 7 DOWNTO 0 DO
            IF (LONGINT(crc) < 0) THEN
               crc := (crc Shl 1) XOR polycrc32a
            ELSE
               crc := crc Shl 1;
         crc32abyfb := crc;
         END; {crc32abyfb}



      TYPE
         WordArray = ARRAY[ 0..32760 ] of WORD;

      CONST
         RandCtDispPdPwr = 3;  { 3 => 2**3 => 8 => mask = 7 }
         RandCtDispMask = PRED( 1 Shl RandCtDispPdPwr );


      FUNCTION GetRandWds( VAR buf; numwds: WORD ): BOOLEAN;
         VAR
            randwdar: WordArray ABSOLUTE buf;
            i, t, q: WORD;
            ch: CHAR;
            crc: Lwd;
            crclsby: BYTE ABSOLUTE crc;
            abortct: BYTE;
         BEGIN
         GetRandWds := FALSE;
         TimerInit;

         { DO NOT initialize values; accept whatever is there! }
         IF (crc = 0) THEN  Dec( crc );   { except zero }

         abortct := 0;
         WRITE( Output, ^M^J' Please type for ', numwds, ' random values: ',
                                                 numwds:4, ' to go. ' );

         FOR i := PRED(numwds) DOWNTO 0 DO
            BEGIN

            { restricted window into key acceptance }
            {    display might otherwise provide timing information }
            IF ((i AND RandCtDispMask) = 0) THEN
               WRITE( Output, ^H^H^H^H^H^H^H^H^H^H^H^H, SUCC(i):4, ' to go. ' );

               REPEAT
               { we want HUMAN timing:  Flush pending keys! }
               WHILE KeyPressed DO  ch := ReadKey;

                  REPEAT
                  { get 16-bit count (rolls over 18.2 times/sec) }
                  t := GetCount;

                  { confuse count value }
                  t := t XOR (t Shr 1);
                  t := t XOR (t Shl 2);

                  { encrypt each count byte }
                  crc := crc32abyfb( crc, Hi(t) );
                  q := crclsby;
                  crc := crc32abyfb( crc, Lo(t) );
                  q := WORD( Swap(q) + crclsby );

                  UNTIL KeyPressed;

               { get char and check for user abort }
               ch := ReadKey;  IF (ch = #0) THEN  ch := ReadKey;
               IF (ch = ^[) THEN  Inc( abortct )  ELSE  abortct := 0;
               IF (abortct >= 3) THEN  BEGIN  TimerRecover;  Exit;  END;

               { skip keyboard repeat and few-key alternations }
               UNTIL NewChar( ch );

            { new char => save encrypted count as random data }
            randwdar[i] := q;
            END;

         WRITE( Output, ^M^J' DONE.  Please press Esc: ' );
         REPEAT  UNTIL KeyPressed AND (Readkey = ^[);

         TimerRecover;
         GetRandWds := TRUE;
         END; {GetRandWds}


      { end of really-random generator }



      { start protected additive RNG }


      TYPE
         LwdArray = ARRAY[ 0..16380 ] of Lwd;
         LwdArrayPtr = ^LwdArray;

      TYPE
         WordBitList = RECORD
                       MaxLen, Len: WORD;
                       Wdat: ARRAY[ 1..32760 ] of WORD;
                       END;
         WordBitListPtr = ^WordBitList;




      CONST  { some mnemonics for inline machine code }
         AdcDXBXSIoff = $5013;   AddAXBXSIindFORCE = $0003;
         AddBXBP = $dd03;   CmpBXCX = $d93b;   IncDI = $47;
         Jmp = $eb;   Jb = $72;   Jnz = $75;   OrBXBX = $db0b;
         MovAXDS = $d88c;   MovBXBP = $eb89;   MovBXDIind = $1d8b;
         MovBXSIindAXFORCE = $0089;   MovBXSIoffDX = $5089;
         PopBP = $5d;   PopCX = $59;   PopDI = $5f;
         PopDS = $1f;   PopES = $07;   PopSI = $5e;
         PushAX = $50;   PushBX = $53;
         SegES = $26;   ShlBX1 = $e3d1;   SubBXCX = $d92b;
         XorAXAX = $c033;   XorDXDX = $d233;



{$IFNDEF InlRNG }

      FUNCTION LwdTabRNGaddStep( wbl: WordBitListPtr;
                deltab: POINTER; TabMod, curofs: WORD ): Lwd;
         { Additive RNG for 32-bit data and general polynomial }
         VAR
            temp: Lwd;
            i: BYTE;
            el: WORD;
         BEGIN
         temp := 0;
         WITH wbl^ DO
            FOR i := 1 TO PRED(Len) DO
               BEGIN
               el := Wdat[i] + curofs;
               IF (el >= TabMod) THEN  Dec( el, TabMod );
               Inc( temp, LwdArrayPtr(deltab)^[ el ] );
               END;
         LwdArrayPtr(deltab)^[curofs] := temp;
         LwdTabRNGaddStep := temp;
         END; {LwdTabRNGaddStep}

{$ELSE}

      FUNCTION LwdTabRNGaddStep( wbl: WordBitListPtr;
                deltab: POINTER; TabMod, curofs: WORD ): Lwd; Inline(
         MovBXBP/ PopBP {curofs}/ PopCX {TabMod}/ MovAXDS/ PopSI/PopDS {deltab}/
         PopDI/PopES {wbl}/ PushAX {DS}/ PushBX {BP}/
         IncDI/IncDI/IncDI/IncDI {wbl.wdat[1]}/ XorDXDX/XorAXAX {result}/ Jmp/17 /
              AddBXBP/ CmpBXCX/ Jb/2 / SubBXCX/ {now have table el #}
              ShlBX1/ShlBX1 {4 bytes per}/ >AddAXBXSIindFORCE/ AdcDXBXSIoff/2/
            SegES/MovBXDIind {poly el #}/ IncDI/IncDI/ OrBXBX/ Jnz/<-26 /
         MovBXBP/ ShlBX1/ShlBX1/ >MovBXSIindAXFORCE/MovBXSIoffDX/2 /
         PopBP/ PopDS );

{$ENDIF}



      TYPE
         WBL15 = ARRAY[ -1..15 ] of WORD;

      CONST  { deg 31 PRIMITIVE mod 2 15-nomials for state expansion }
         { these particular polynomials (c) Copyright 1990, T. F. Ritter }
         wbl31a: WBL15 = (15,15,  31,27,24,23,20,17,13,11,10,9,6,5,2,1,0 );
         wbl31b: WBL15 = (15,15,  31,29,28,27,25,20,16,11,8,7,5,4,2,1,0 );


      TYPE
         WBL9 = ARRAY[ -1..9 ] of WORD;

      CONST  { large PRIMITIVE mod 2 9-nomials for state expansion }
         { these particular polynomials (c) Copyright 1990, T. F. Ritter }
         wbl127a: WBL9 = (9,9,  127,109,75,72,64,51,50,38,0 );
         wbl127b: WBL9 = (9,9,  127,113,111,84,77,65,32,7,0 );
         wbl607a: WBL9 = (9,9,  607,589,477,335,241,175,158,134,0 );
         wbl2281a: WBL9 = (9,9,  2281,2264,1845,1500,1439,802,401,345,0 );

      CONST  { large PRIMITIVE mod 2 9-nomials for sequence generation }
         { these particular polynomials (c) Copyright 1990, T. F. Ritter }
         wbl1279a: WBL9 = (9,9, 1279,1006,903,846,585,406,355,269,0 );
         wbl4253a: WBL9 = (9,9,  4253,3803,3797,2754,1456,1319,965,827,0 );
         wbl11213a: WBL9 = (9,9, 11213,10743,10485,8051,6255,4262,2638,2516,0 );

      { An RNG 32 bits wide and 11213 elements long contains 358,816 bits! }
      {    That is almost 44K of RNG state; 11,213x the typical 32-bit LCG. }



      CONST
         RNGlastEl = 12000;  { support polys through deg 11213 }
         Lwd31fillSize = 32 * 4;   { 128 }
         Lwd31initSize = 31 * 4;   { 124 }

      VAR  { static memory allocation: RNG storage }
         RNGlwdTab: ARRAY[ 0..RNGlastEl ] of Lwd;  { 48004 bytes }
         polyp: WordBitListPtr;
         curofs: WORD;


      FUNCTION LwdAddRNG: Lwd;
         { a full 32-bit Additive RNG step using polyp }
         BEGIN
         LwdAddRNG :=
            LwdTabRNGaddStep( polyp, @RNGLwdTab, SUCC(RNGlastEl), curofs );
         IF (curofs = 0) THEN  curofs := RNGlastEl  ELSE  Dec( curofs );
         END; {LwdAddRNG}


      PROCEDURE ResetRNG;
         { RNG runs backwards: next element after init is RNGlastEl }
         {    init of els 0..31 generates -1, -2, etc. (mod size) }
         BEGIN
         curofs := RNGlastEl;
         END; {ResetRNG}


      PROCEDURE CheckRNGinit;
         { guarantee at least one lsb '1' in init }
         VAR
            i: BYTE;
            comb: Lwd;
         BEGIN
         comb := 0;
         FOR i := 0 TO 31 DO
            comb := comb OR RNGlwdTab[i];
         IF NOT ODD(comb) THEN
            Inc( RNGlwdTab[0] );
         END; {CheckRNGinit}



      PROCEDURE Extend31to127( p: WordBitListPtr );
         { extend 31-element RNG init to 127 elements }
         VAR
            i: BYTE;
            l: Lwd;
         BEGIN
         CheckRNGinit;
         ResetRNG;
         polyp := @wbl31a;
         FOR i := 32 TO 127 DO  l := LwdAddRNG;
         polyp := p;
         END; {Extend31to127}


      PROCEDURE Extend31toX( x: WORD; p: WordBitListPtr );
         { extend 31-element RNG init to at least 2281 elements }
         VAR
            i, j: WORD;
            l: Lwd;
         BEGIN
         CheckRNGinit;
         ResetRNG;
         polyp := @wbl31b;
         FOR i := 32 TO 127 DO  l := LwdAddRNG;
         WRITE( Output, i:7 );

         polyp := @wbl127b;
         FOR i := 128 TO 607 DO  l := LwdAddRNG;
         WRITE( Output, ^H^H^H^H^H^H^H, i:7 );

         polyp := @wbl607a;
         j := 2281;
         IF (x < j) THEN  j := x;
         FOR i := 608 TO j DO  l := LwdAddRNG;
         WRITE( Output, ^H^H^H^H^H^H^H, i:7 );

         polyp := @WBL2281a;
         FOR i := 2282 TO x DO  l := LwdAddRNG;
         WRITE( Output, ^H^H^H^H^H^H^H, i:7 );

         polyp := p;
         END; {Extend31toX}



      FUNCTION UserKeyRNGinit: BOOLEAN;
         { get user key string, transform into RNG init, }
         {    extend through deg 127, RNG cipher with buf data, }
         {    move buf into RNG init, extend as needed }
         VAR
            s: STRING;
         BEGIN
         UserKeyRNGinit := FALSE;

         IF NOT GetUserKey( s ) THEN
            BEGIN  WRITE( Output, ^M^J' Unable to get user key. ' );  Exit;  END;

         { transform user key into 992-bit deg 31 RNG init }
         UserKeyCRCinit( RNGlwdTab, poly31ar, s[1], Length(s) );
         FillChar( s, SizeOf(s), 0 );   { hide plaintext key }
         Pack31bitLwd32AR( RNGlwdTab );   { deg 31 RNG init }

         Extend31To127( @wbl127a );   { now have deg 127 RNG }

         UserKeyRNGinit := TRUE;
         END; {UserKeyRNGinit}



      TYPE
         CiLevTy = ( Low, Med, High, Cust );


      PROCEDURE Init31toCiLev( VAR x; lev: CiLevTy );
         { copy x to deg 31 RNG and extend as needed }
         BEGIN
         WRITE( Output, ^M^J^J' Starting initialization: ' );
         Move( x, RNGlwdTab, Lwd31initSize );

            CASE lev OF
            Low: Extend31ToX( 1279, @wbl1279a );
            Med: Extend31ToX( 4253, @wbl4253a );
            High: Extend31ToX( 11213, @wbl11213a );
            END; {case}

         END; {Init31ToCiLev}




      CONST  { static memory allocation: jitterizer globals }
         takect: BYTE = 1;
         LwdValOff: Lwd = 0;


      FUNCTION LwdJitterOffRNG: Lwd;
         { Jitterize and Value Offset an RNG sequence }
         {    hide RNG sequence from cryptanalysis }
         VAR
            i: BYTE;
            x: Lwd;
         BEGIN
         { skip avg of 2.5 rands, then take avg of 32.5 rands }
         {   we drop about 7% of the original sequence }

         IF (takect > 0) THEN
            BEGIN
            LwdJitterOffRNG := LwdValOff + LwdAddRNG;  { offset }
            Dec( takect );
            Exit;
            END;

         LwdValOff := LwdAddRNG;  { skip once; set skip count }

         takect := LwdValOff AND $ff;
         FOR i := 0 TO (takect AND 3) DO  { skip 0..3 more }
            LwdValOff := LwdAddRNG;  { offset is the last skip }

         LwdJitterOffRNG := LwdValOff + LwdAddRNG;  { take once }
         takect := takect Shr 2;  { take 0..63 more later }

         END; {LwdJitterOffRNG}


      { end protected additive RNG }



      { start dynamic substitution }


      CONST
         Cld = $fc;  Lodsb = $ac;
         MovAHDIind = $258a;  MovBXDS = $db8c;  MovDSBX = $db8e;
         MovSIoffAH = $6488;  Stosb = $aa;


{$IFNDEF InlExg }

      PROCEDURE ExchangeChars( VAR x, y: CHAR );
         VAR
            t: CHAR;
         BEGIN
         t := x;  x := y;  y := t;
         END; {ExchangeChars}

{$ELSE}

      PROCEDURE ExchangeChars( VAR x, y ); Inline( MovBXDS /
         PopDI/PopES {y}/ PopSI/PopDS {x}/ Cld/
         SegES/MovAHDIind/ Lodsb/ Stosb/ MovSIoffAH/<-1 / MovDSBX );

{$ENDIF}


      TYPE
         SubBox = ARRAY[ CHAR ] of CHAR;
         DynSubBox = RECORD fwd, inv: SubBox; END;

      CONST
         LwdBoxLastEl = PRED( SizeOf(SubBox) Shr 2 );  { to Lwd's }

      TYPE
         LwdBox = ARRAY[ 0..LwdBoxLastEl ] of Lwd;  { same size box }



      PROCEDURE DynSubInit( VAR f, randbox: SubBox );
         { init substitution table }
         VAR
            ch: CHAR;
         BEGIN
         FOR ch := #0 TO #255 DO  f[ ch ] := ch;
         FOR ch := #0 TO #255 DO
            ExchangeChars( f[ ch ], f[ randbox[ch] ] );
         END; {DynSubInit}


      FUNCTION DynSubF( VAR f: SubBox; xi, rand: CHAR ): CHAR;
         { encipher a character }
         BEGIN
         DynSubF := f[xi];
         ExchangeChars( f[xi], f[rand] );
         END; {DynSubF}


      PROCEDURE InvDynSubInit( VAR f: DynSubBox; randbox: SubBox );
         { init substitution and inverse substitution tables }
         VAR
            ch: CHAR;
         BEGIN
         DynSubInit( f.fwd, randbox );
         FOR ch := #0 TO #255 DO  f.inv[ f.fwd[ ch ] ] := ch;
         END; {InvDynSubInit}


      FUNCTION InvDynSubF( VAR f: DynSubBox; yi, rand: CHAR ): CHAR;
         { decipher a character }
         VAR
            j: CHAR ABSOLUTE rand;
            xi, yj: CHAR;
         BEGIN
         xi := f.inv[yi];
         yj := f.fwd[j];
         ExchangeChars( f.inv[yi], f.inv[yj] );
         ExchangeChars( f.fwd[xi], f.fwd[j] );
         InvDynSubF := xi;
         END; {InvDynSubF}



      CONST  { set size of dynsub array }
         DSBApwr = 4;   { 4 => LastDSBAel = 15 }
         LastDSBAel = PRED( 1 Shl DSBApwr );

      VAR  { static memory allocation: dynsub structures }
         { NOTE:  f2 is an ARRAY of dynamic substitution boxes }
         f1: DynSubBox;
         f2: ARRAY[ 0..LastDSBAel ] of DynSubBox;


      PROCEDURE DDynSubInit;
         { init array of DynSub boxes, plus first DynSub }
         VAR
            i, j: BYTE;
            sb: SubBox;
            lb: LwdBox ABSOLUTE sb;
         BEGIN
         FOR i := 0 TO LastDSBAel DO
            BEGIN
            FOR j := 0 TO LwdBoxLastEl DO  lb[j] := LwdJitterOffRNG;
            DynSubInit( f2[i].fwd, sb );
            END;
         FOR j := 0 TO LwdBoxLastEl DO  lb[j] := LwdJitterOffRNG;
         DynSubInit( f1.fwd, sb );
         END; {DDynSubInit}


      FUNCTION DDynSubF( d: CHAR; t: Lwd ): CHAR;
         { double dynamic substition of d }
         VAR
            qch: ARRAY[ 0..3] of CHAR ABSOLUTE t;
         BEGIN
         d := DynSubF( f1.fwd, d, qch[1] );
         d := DynSubF( f2[ ORD(qch[2]) AND LastDSBAel ].fwd, d, qch[3] );
         DDynSubF := d;
         END; {DDynSubF}



      PROCEDURE DInvDynSubInit;
         { init array of InvDynSub boxes, plus first InvDynSub }
         VAR
            i, j: BYTE;
            sb: SubBox;
            lb: LwdBox ABSOLUTE sb;
         BEGIN
         FOR i := 0 TO LastDSBAel DO
            BEGIN
            FOR j := 0 TO LwdBoxLastEl DO  lb[j] := LwdJitterOffRNG;
            InvDynSubInit( f2[i], sb );
            END;
         FOR j := 0 TO LwdBoxLastEl DO  lb[j] := LwdJitterOffRNG;
         InvDynSubInit( f1, sb );
         END; {DInvDynSubInit}


      FUNCTION DInvDynSubF( d: CHAR; t: Lwd ): CHAR;
         { double inverse dynamic substitution of d }
         VAR
            qch: ARRAY[ 0..3] of CHAR ABSOLUTE t;
         BEGIN
         d := InvDynSubF( f2[ ORD(qch[2]) AND LastDSBAel ], d, qch[3] );
         d := InvDynSubF( f1, d, qch[1] );
         DInvDynSubF := d;
         END; {DInvDynSubF}


      { end dynamic substitution }



      PROCEDURE XorEncrLwds( VAR buf; lwds: WORD );
         { xor-encipher memory data with RNG }
         VAR
            lwbuf: LwdAR ABSOLUTE buf;
            i: WORD;
         BEGIN
         FOR i := 0 TO PRED(lwds) DO
            lwbuf[i] := lwbuf[i] XOR LwdJitterOffRNG;
         END; {XorEncrLwds}



      { start cipher file structure }


      TYPE
         CipherHeader = RECORD
            EnciFlag: Lwd;
            CipherMfr: WORD;
            CiphTy: BYTE;
            RevLev: BYTE;
            ChanTag: Lwd;
            END; {record}

      CONST
         SizCHdr = SizeOf( CipherHeader );
         PSizCHdr = PRED( SizCHdr );

      TYPE
         CHdrChAR = ARRAY[ 0..PSizCHdr ] of CHAR;

      CONST
         Enciphered = $a6c3f012;
         BJSoftware = 1;
         DDynSub = 1;
         RevMask = $fc;  { 1111 11xx = 0..$3f Shl 2 }
         LevMask = 3;    { xxxx xx11 = 0..3 }
         Rev01 = 1 Shl 2;   { 01 = 0000 0100 }
            { only INCOMPATIBLE changes imply a different rev }

      CONST  { static memory allocation: Current Header }
         HeadConst: CipherHeader = (
            EnciFlag: Enciphered;
            CipherMfr: BJSoftware;   { BJS compatible if msb = 0 }
            CiphTy: DDynSub;
            RevLev: Rev01;   { RNG Lev set dynamically }
            ChanTag: 0 );

      CONST
         MsgKeyFill = Lwd31fillSize;  { 32 Lwds * 4 by/Lwd = 128 by }
         MsgKeyLen = Lwd31initSize;  { 31 Lwds * 4 by/Lwd = 124 by }
         PredMsgKeyFill = PRED( MsgKeyFill );


      { end cipher file structure }



      { start file operations }


      CONST  { set size of file buffer }
         fbufpwr = 12;   { 9 => lastfbufel = 511; 12 => 4095 }
         lastfbufel = PRED( 1 Shl fbufpwr );

      VAR  { static memory allocation }
         fromfi, tofi: FILE;
         filebuf: ARRAY[ 0..lastfbufel ] of CHAR;

      CONST  { set progress display period }
         ProgDispPwr = 8;  { 8 => 2**8 = 256 => mask = 255 }
         ProgDispMask = PRED( 1 Shl ProgDispPwr );



      PROCEDURE encipher( VAR data: CHdrChAR; byct: WORD; lev: CiLevTy );
         { read, encipher, write buffer, until done }
         VAR
            i, got, did: WORD;
            tot: Lwd;
            tempMK: ARRAY[ 0..PredMsgKeyFill ] of BYTE;
         BEGIN
         tot := 0;

         { set RNG level in header global and move to filebuf[ 0..11 ] }
         Inc( HeadConst.RevLev, ORD(lev) );
         Move( HeadConst, filebuf, SizCHdr );
         got := SizCHdr;

         { get user key and use it to init deg 127 RNG }
         WRITE( Output,
^M^J' A strong user key is LONGER than 20 or 30 characters and has an ',
^M^J'    ALTERED spelling, capitalization, punctuation and/or spacing. ' );
         IF NOT UserKeyRNGinit THEN  Exit;


         { make 124-byte message key in tempMK }
         WRITE( Output,
^M^J^J' Generating message key (type Esc-character 3 times to abort): ',
  ^M^J'    Please type different keys at random until counter reaches 0. ' );
         IF NOT GetRandWds( tempMK, MsgKeyFill Shr 1 ) THEN
            BEGIN  WRITE( Output, ^M^J^J' Unable to make message key. ' );  Exit;  END;
         Pack31bitLwd32AR( tempMK );

         { copy message key to filebuf[ 12..135 ] and encipher it there }
         Move( tempMK, filebuf[got], MsgKeyLen );
         XorEncrLwds( filebuf[got], MsgKeyLen Shr 2 );   { cipher with RNG }
         Inc( got, MsgKeyLen );

         { unenciphered message key inits RNG }
         Init31toCiLev( tempMK, lev );
         FillChar( tempMK, SizeOf(tempMK), 0 );   { hide plain message key }


         DDynSubInit;
         WRITELN( Output, ^M^J' Starting data enciphering: ' );

         { encipher passed data to filebuf[ 136..147 ] }
         FOR i := 0 TO PRED(byct) DO
            filebuf[got+i] := DDynSubF( data[i], LwdJitterOffRNG );
         Inc( got, byct );

         BlockWrite( tofi, filebuf, got, did );
         IF (did <> got) THEN
            BEGIN  WRITE( Output, ^M^J' Unable to write file. ' );  Exit;  END;

            REPEAT
            Inc( tot, did );  WRITE( Output, ^M' ', tot );
            BlockRead( fromfi, filebuf, SizeOf(filebuf), got );
            IF (got = 0) THEN  Exit;  { done or read failure }
            FOR i := 0 TO PRED(got) DO
               BEGIN
               IF ((i AND ProgDispMask) = 0) THEN
                  WRITE( Output, ^M' ', tot + i );
               filebuf[i] := DDynSubF( filebuf[i], LwdJitterOffRNG );
               END;
            BlockWrite( tofi, filebuf, got, did );
            UNTIL (did <> got);  { write failure }

         END; {encipher}



      PROCEDURE decipher( CMfr: WORD; CTy, RLev: BYTE );
         { read, decipher, write buffer, until done }
         VAR
            i, got, did: WORD;
            rev: BYTE;
            lev: CiLevTy;
            tot: Lwd;
         BEGIN
         tot := 0;

         IF (INTEGER(CMfr) < 0) THEN
            BEGIN
            WRITE( Output, ^M^J' Incompatible Mfr ', CMfr );
            Exit;
            END;

         IF (CTy <> DDynSub) THEN
            BEGIN
            WRITE( Output, ^M^J' Unknown cipher ', CTy );
            Exit;
            END;

         rev := (RLev AND RevMask);
            CASE rev OF
            Rev01: ;
            ELSE {case}
               WRITE( Output, ^M^J' Unknown Rev ', rev Shr 2 );
               Exit;
            END; {case}

         lev := CiLevTy( RLev AND LevMask );
         IF (lev = Cust) THEN
            BEGIN
            WRITE( Output, ^M^J' Unable to decipher Custom RNG Lev ' );
            Exit;
            END;


         { read 124-byte message key }
         BlockRead( fromfi, filebuf, MsgKeyLen, got );
         IF (got < MsgKeyLen) THEN
            BEGIN  WRITE( Output, ^M^J' Unable to read file. ' );  Exit;  END;

         { user key deciphers message key }
         IF NOT UserKeyRNGinit THEN  Exit;
         XorEncrLwds( filebuf, MsgKeyLen Shr 2 );   { cipher with RNG }

         { deciphered message key inits RNG }
         Init31toCiLev( filebuf, lev );


         DInvDynSubInit;
         WRITELN( Output, ^M^J' Starting data deciphering: ' );

            REPEAT
            BlockRead( fromfi, filebuf, SizeOf(filebuf), got );
            IF (got = 0) THEN  Exit;  { done or read failure }
            FOR i := 0 TO PRED(got) DO
               BEGIN
               IF ((i AND ProgDispMask) = 0) THEN
                  WRITE( Output, ^M' ', tot + i );
               filebuf[i] := DInvDynSubF( filebuf[i], LwdJitterOffRNG );
               END;
            BlockWrite( tofi, filebuf, got, did );
            Inc( tot, did );  WRITE( Output, ^M' ', tot );
            UNTIL (did <> got);  { write failure }

         END; {decipher}



      TYPE
         Str127 = STRING[ 127 ];

      VAR
         CmdLinP: ^Str127;


      PROCEDURE Action;
         { select and invoke operation, then clean up }
         VAR
            got: WORD;
            lev: CiLevTy;
            EnciSw: BOOLEAN;
            Header: CipherHeader;
         BEGIN

         EnciSw := TRUE;

         BlockRead( fromfi, Header, SizCHdr, got );
         IF (got = 0) THEN
            BEGIN  WRITE( Output, ^M^J' Unable to read file. ' );  Exit;  END;
         IF (got = SizCHdr) THEN
            { enciphering if file does not start w/ enciphered flag }
            EnciSw := (Header.EnciFlag <> Enciphered);

         { command line can force enciphering of enciphered files }
         IF (Pos('/e',CmdLinP^) <> 0) THEN  EnciSw := TRUE;

         { command line can select data enciphering level }
         lev := Med;
         IF (Pos('/l',CmdLinP^) <> 0) THEN  lev := Low
         ELSE
            IF (Pos('/h',CmdLinP^) <> 0) THEN  lev := High;

         IF EnciSw THEN
            encipher( CHdrChAR(Header), got, lev )
         ELSE
            WITH Header DO
               decipher( CipherMfr, CiphTy, RevLev );

         { done:  destroy sensitive memory data }
         FillChar( Header, SizeOf(Header), 0 );
         FillChar( filebuf, SizeOf(filebuf), 0 );
         FillChar( RNGlwdTab, SizeOf(RNGlwdTab), 0 );
         FillChar( f1, SizeOf(f1), 0 );
         FillChar( f2, SizeOf(f2), 0 );
         END; {Action}



      PROCEDURE OpenClose;
         { open files, call action, close files }
         BEGIN
         ASSIGN( fromfi, '' );  { StdIn }
         RESET( fromfi, 1 );
         IF (IOresult = 0) THEN
            BEGIN
            ASSIGN( tofi, '' );  { StdOut }
            REWRITE( tofi, 1 );
            IF (IOresult = 0) THEN
               BEGIN
               Action;
               CLOSE( tofi );
               END;
            CLOSE( fromfi );
            END;
         WRITELN( Output, ^M^J' DONE. ' );
         END; {OpenClose}


      { end file operations }



      PROCEDURE lcST( VAR s: STRING );
         { convert string to lower case }
         VAR
            i: BYTE;
         BEGIN
         FOR i := 1 TO Length(s) DO
            CASE s[i] OF
            'A'..'Z': Inc( s[i], ORD('a') - ORD('A') );
            END; {case}
         END; {lcST}



      BEGIN
      Ownership;

      CmdLinP := Ptr( PrefixSeg, $80 );
      lcST( CmdLinP^ );

      IF (Pos('/o', CmdLinP^) <> 0) OR
         (Pos('?', CmdLinP^) <> 0) THEN
         ShowOptions
      ELSE
         BEGIN
         OpenClose;
         Advert;
         END;

      END.


      { end file cloak.pas }
